'' System Version 2.003
'' loaderInit Changes:
''
''   2006/09/24 - Corrected action table for i2cRead0Cmd
''   2006/10/02 - Improved assembly comments.  Changed i2cReset, i2cStop
''   2006/11/02 - Changed data setup time.  Changed i2cReset timing
''   2006/11/03 - Changed read/write method speed to 100KHz
''   2006/11/04 - Added checkPresence, writeWait, and computeTimes methods
''   2006/11/06 - Limited boot loading to 32K - 16 (OS uses last 16 bytes)
''                Added ioVerifyCmd and verifyEEPROM method
''   2006/11/09 - Modified boot and verify to use the minimum of the actual
''                length of the program stored in vbase ($0008) or the
''                specified byte count.
''                Note also that these routines don't know about option bits.
''   2006/11/10 - Now control block address is passed to start routine
''   2007/01/09 - Added SPI routines for SD card FAT access
''   2007/01/13 - Corrected checksum when ioBoot or ioVerify
''   2007/01/13 - Stores stack marker & clears VAR area on ioBootCmd
''   2007/02/06 - Stores stack marker & clears VAR area on ioSpiBoot
''   2007/02/13 - Changed the way verify mode was done, combined code
''   2007/02/20 - Added ioStopLdr option to stop the loader's cog
''   2007/02/22 - Corrected bootSDCard.  Needs start and initSDCard calls first.

'' This portion is normally run only once during initialization and the driver remains
'' resident in a cog.  These routines can be used completely independently of the rest
'' of the Propeller OS.  The start routine here expects the address of a 2 long area
'' to be used for passing information to the I/O routines in the COG.  This area should
'' be located in an area of memory not expected to be overlaid by data or a program that
'' might be loaded since the COG routines will be accessing this information after an
'' operation has completed.

'' This object provides an I2C EEPROM read/write routine that can handle both 100KHz and
'' 400KHz bus speeds and EEPROM page sizes of 64, 128, or 256 bytes (or no paging/no delay
'' as with Ramtron serial RAM).  The SPIN interpreter can be started after reading, either
'' in the same COG used by these routines or in a free COG.  The control information is
'' passed in a 2 long parameter block whose address is passed to the COG when it is started.
'' The parameter block is updated when the operation is completed.  Note that these are shown
'' here as they appear in a long value rather than the order of the bytes in memory.

'' -------------------------------------------------------------------
'' |   cmd/status   |          I/O pin / device / address            |
'' -------------------------------------------------------------------
'' |           byte count           |          HUB address           |
'' -------------------------------------------------------------------

'' The EEPROM address is in the same format used by other routines with the I/O pin pair
'' in bits 21..19, the device address in bits 18..16, and the 64K address in bits 15..0.
'' Note that the I/O pin pair is the number of the SCL pin divided by 2.  The SDA pin is
'' always the next higher numbered pin.  The command code is in the low order bits of the
'' high order byte of the first long (see ioCmdMask).  This is always non-zero to indicate
'' that a command is to be performed by the COG routines.  When the command is finished,
'' this is set to zero.  The errorFlag bit is set to one if a NAK was read after a write
'' transfer.  This is the only error reported by these routines.  A read operation and
'' zero-length writes do involve several write transfers for addressing, but the data
'' read transfer has no error checking.  When the command is completed, the device address,
'' byte count, and HUB address are all updated to their values at that time.  For the
'' verify operation (ioVerifyCmd), an error is reported if the checksum is not zero and
'' the HUB address field is not incremented.  It may be used for some other checksum
'' reporting in the future.

'' The pins used for the boot EEPROM I2C bus (at least on Parallax's Demo Board) do not
'' have a pullup on SCL.  This requires that SCL be driven both high and low.  If the bus
'' used is on pins 28 and 29, SCL is actively driven at all times.

'' These EEPROM read/write routines do not provide for waiting for the write to complete
'' nor do they check for paged writes.  All bytes in a multi-byte write must lie within
'' a single EEPROM page since the EEPROM write address counter wraps around at a page
'' boundary.  Similarly, for multi-byte reads, all requested bytes must lie within the
'' same device since the sequential read counter wraps around at the device boundary.

'' Command codes are provided for devices with zero, one, or two address bytes following
'' the device selection byte.  As for all I2C devices, addressing is done using write
'' mode and the device is reselected in read mode after the last address byte.  In the
'' case of ioRead0Cmd, the device is initially selected in read mode.  For 8-bit addresses,
'' the device select code is taken from bits 15-8 of the address value.  For the case
'' without address bytes, the device select code is taken from bits 7-0 of the address value.
'' These device select codes must have their least significant bit set to zero (for write
'' mode) except in the case of ioRead0Cmd where it must be set to one for proper operation.

'' SPI data is handled a little differently.  For ioSpiInit, the 6 bit pin numbers for DO,
'' Clk, DI, and CS are given from MSB to LSB of the 24 bit address field of the command and
'' are used for all further I/O operations (until an ioSpiStop is done).

CON
'' Command code and error information for I2C driver
'' (For convenience in using just OS_loaderInit, these are included here.  The "master"
''  copies are considered the ones in OS_loader and these must be kept up-to-date).

  ioReadCmd   = %00000001                      ' Read from EEPROM to HUB RAM (16 bit addresses)
  ioWriteCmd  = %00000010                      ' Write to EEPROM from HUB RAM (16 bit addresses)
  ioRead1Cmd  = %00000011                      ' Read from a device with only 8-bit addresses
  ioWrite1Cmd = %00000100                      ' Write to a device with only 8-bit addresses
  ioRead0Cmd  = %00000101                      ' Read from a device without address bytes
  ioWrite0Cmd = %00000110                      ' Write to a device without address bytes
  ioBootCmd   = %00001000                      ' Read from EEPROM to HUB RAM, then start a
                                               '  new SPIN interpreter in the COG whose ID is
                                               '  supplied in the lower 3 bits of this command
                                               '  This COG is stopped before the read is done
                                               '  unless it's the one used to execute the loader
  ioSpiInit   = %00010000                      ' Initialize the specified SPI bus and SD card
  ioSpiStop   = %00010001                      ' Change all SD card pins to inputs
  ioSpiRead   = %00010010                      ' Read one or more bytes from the SD card
  ioSpiWrite  = %00010011                      ' Write one or more bytes from the SD card
  ioSpiBoot   = %00011000                      ' Like ioBootCmd, but uses ioSpiRead for loading
  ioCmdMask   = %00011111                      ' Used to mask off command bits
  ioSpiMask   = %00010000                      ' Used to test for SPI command codes

' Options for commands

  ioNoStore   = %00100000                      ' If set, data is not stored into main memory
                                               '  If ioBootCmd or ioSpiBoot, no cogs are
                                               '   stopped and a new cog is not started.
  ioLowSpeed  = %01000000                      ' If set, I2C runs at 100KHz rather than 400KHz
  ioStopLdr   = %10000000                      ' If set, the loader's cog is stopped after a boot

' Return status

  ioWriteErr  = %10000000                      ' An error occurred during an I2C write (NAK)

  ioTestRdy   = ioCmdMask << 24                ' Used to test 1st control long for ready
  ioTestErr   = ioWriteErr << 24               ' Used to test 1st control long for write error

'' Other constants from OS_loader

  i2cBootSCL  = 28                             ' Boot EEPROM SCL pin
  bootAddr    = i2cBootSCL << 18               ' Address of boot EEPROM
  clkfreqVal  = $0000                          ' Current CLKFREQ value stored here
  clksetVal   = $0004                          ' Current CLKSET value stored here
  chksumVal   = $0005                          ' Checksum over memory stored here
  vbase       = $0008                          ' Length of Spin program loaded (# longs * 4)
  dbase       = $000A                          ' Address of start of stack (marker below)

VAR
  long cog, control



PUB checkPresence(addr) | t
'' This routine checks to be sure there is an I2C bus and an EEPROM at the
'' specified address.  Note that this routine cannot distinguish between a
'' 32Kx8 and a 64Kx8 EEPROM since the 16th address bit is a "don't care"
'' for the 32Kx8 devices.  Return true if EEPROM present, false otherwise.
  t := ioReadCmd | ioLowSpeed
  ' rdy
  long[control][1] := 0                                ' Attempt to address the device
  long[control][0] := (t << 24) | (addr & $FFFFFF)
  rdy
  return (long[control][0] & ioTestErr) == 0           ' Return false on error

{{
PUB writeWait(addr) | t                                '' Wait for EEPROM to complete write
  t := cnt
  repeat until checkPresence(addr)                     ' Maximum wait time is 20ms
    if (cnt - t) > (clkfreq / 50)
      return true                                      ' Return true if a timeout occurred
  return false                                         ' Otherwise return false
}}
PUB computeTimes(loadBuf) | t                             '' Set up timing constants in assembly
  loadBuf+=1852
  t:=clkfreq / 10000                                               '  (Done this way to avoid overflow)
  long[loadBuf][0] := (t *  6) / 1000   ' Data setup time -  600ns (100KHz)
  long[loadBuf][1] := (t * 47) / 1000   ' Clock low time  - 4700ns (100KHz)
  long[loadBuf][2] := (t * 4) / 100   ' Clock high time - 4000ns (100KHz)
  long[loadBuf][3] := (t *  25) / 10000   ' Data setup time -  250ns (400KHz)
  long[loadBuf][4] := (t * 13) / 1000   ' Clock low time  - 1300ns (400KHz)
  long[loadBuf][5] := t / 100   ' Clock high time - 1000ns (400KHz)
  long[loadBuf][6] := clkfreq / 100000                      ' Pause between checks for operations

pub rdy
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish

PUB initSDCard(DO,Clk,DI,CS)                        '' Initialize SD card access
  ' rdy
  long[control][1] := 0
  long[control][0] := ioSpiInit << 24 | DO << 18 | Clk << 12 | DI << 6 | CS
  rdy
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB stopSDCard                                         '' Stop SD card access
  ' rdy
  long[control][1] := 0
  long[control][0] := ioSpiStop << 24
  rdy
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB readSDCard(addr,buffer,count)                      '' Read block(s) from SD card to RAM
  ' rdy
  long[control][1] := (count << 16) | (buffer & $FFFF)
  long[control][0] := (ioSpiRead << 24) | (addr & $FFFFFF)
  rdy
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB writeSDCard(addr,buffer,count)                     '' Write block(s) to SD card from RAM
  ' rdy
  long[control][1] := (count << 16) | (buffer & $FFFF)
  long[control][0] := (ioSpiWrite << 24) | (addr & $FFFFFF)
  rdy
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB bootSDCard(addr,count,watch) | t                   '' Boot from an SD card
  if count < 16                                        ' Must load at least 16 bytes
    return true
  repeat t from 0 to 7                                 ' Stop all COGs except this one and
    if (t <> cogid) and (t <> (cog-1)) and (t<>watch)  '  the one with the I2C/SPI driver
        cogstop(t)
  t := ioSpiBoot | ioStopLdr | cogid                   ' Tell the SD card driver to load
  long[control][1] := count << 16                      '  into HUB RAM after stopping
  long[control][0] := (t << 24) | (addr & $FFFFFF)     '   this calling cog
  rdy
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB start(ctrlBlk,loadBuf) | t                                 '' Start the I2C I/O driver (standalone)
  control := ctrlBlk                                   ''  using address of 2 longs for control
  ' stop                                                 ' Stop a previous copy
  computeTimes(loadBuf)
  long[control][0] := 0                                ' Initialize the control block
  long[control][1] := 0
  ' cog := cognew(@i2cEntryPoint,control) + 1          ' Start a new cog with the I2C driver
  return control ' cog > 0                             ' Indicate success
{{
PUB setCog(cogNr)
  cog:=cogNr
  
PUB stop                                               '' Stop the I2C I/O driver (standalone)
  if cog > 0
    cogstop(cog - 1)
}}